#ifndef SPCM_CUDA_KERNELS_H
#define SPCM_CUDA_KERNELS_H
/*
**************************************************************************

spcm_cuda_kernels.h                             (c) Spectrum GmbH , 6/2017

**************************************************************************

This file contains miscellaneous CUDA kernel functions.

Feel free to use this source for own projects and modify it in any kind.

**************************************************************************
*/

#ifndef M_PI
#   define M_PI 3.14159265358979323846
#endif

// ----- sorts the channels into continuous arrays for each channel (int16-----
// -- pnIn:     in  Input data. Data sorting as explained in the hardware manual: A0,B0,C0,D0,A1,B1,C1,D1, ...
// -- lNumCh:   in  Number of enabled channels
// -- pnOutCh0: out Data of channel 0
// -- pnOutCh1: out Data of channel 1
// -- pnOutCh2: out Data of channel 2, if lNumCh > 2
// -- pnOutCh3: out Data of channel 3, if lNumCh > 2
// -- pnOutCh4: out Data of channel 4, if lNumCh > 4
// -- pnOutCh5: out Data of channel 5, if lNumCh > 4
// -- pnOutCh6: out Data of channel 6, if lNumCh > 4
// -- pnOutCh7: out Data of channel 7, if lNumCh > 4
__global__ void CudaKernelDemuxChannels (const int16* pnIn, int lNumCh, int16* pnOutCh0, int16* pnOutCh1, int16* pnOutCh2, int16* pnOutCh3, int16* pnOutCh4, int16* pnOutCh5, int16* pnOutCh6, int16* pnOutCh7)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    int lChIdx = i % lNumCh;
    int lSample = i / lNumCh;
    switch (lChIdx)
        {
        // up to 4 channels with linear sorting (M4i)
        // up to 8 channels with linear sorting (M2p)
        case 0: pnOutCh0[lSample] = pnIn[i]; break;
        case 1: pnOutCh1[lSample] = pnIn[i]; break;
        case 2: pnOutCh2[lSample] = pnIn[i]; break;
        case 3: pnOutCh3[lSample] = pnIn[i]; break;
        case 4: pnOutCh4[lSample] = pnIn[i]; break;
        case 5: pnOutCh5[lSample] = pnIn[i]; break;
        case 6: pnOutCh6[lSample] = pnIn[i]; break;
        case 7: pnOutCh7[lSample] = pnIn[i]; break;
        }
    }


// a struct that groups multiple parameters to simplify access in kernels
struct ST_CARDSETUP
    {
    int32 lNumCh;
    int32 lMaxADCValue;
    int32 lIR;
    int32 lLen;
    };

// ----- converts each sample to volt and stores the samples sorted by channel (int16) -----
// -- anSource:     in  The samples as they are generated by the card
// -- afDest:       out The sorted and converted samples
// -- pstCardSetup: in  A struct with some settings that are necessary for the conversion
__global__ void CudaKernelScaleAndDemux (const int16* anSource, cufftReal* afDest, ST_CARDSETUP* pstCardSetup)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    int lChIdx  = i % pstCardSetup->lNumCh;
    int lSample = i / pstCardSetup->lNumCh;
    int o = lChIdx * pstCardSetup->lLen + lSample;
    afDest[o] = (((float)anSource[i]) / pstCardSetup->lMaxADCValue) * ((float)pstCardSetup->lIR) / 1000.0f;
    }

// ----- converts each sample to volt and stores the samples sorted by channel (int16) -----
// -- anSource:     in  The samples as they are generated by the card
// -- afDest:       out The sorted and converted samples
// -- pstCardSetup: in  A struct with some settings that are necessary for the conversion
__global__ void CudaKernelScale (const int16* anSource, cufftReal* afDest, ST_CARDSETUP* pstCardSetup)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    afDest[i] = (((float)anSource[i]) / pstCardSetup->lMaxADCValue) * ((float)pstCardSetup->lIR) / 1000.0f;
    }
__global__ void CudaKernelScale (const int16* anSource, cufftReal* afDest, double dFactor)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    afDest[i] = ((float)anSource[i]) * dFactor;
    }


// CUDA callback for cufftXtSetCallback
__device__ cufftReal CudaCallbackConvertToVolt (void* pvSource, size_t offset, void* pvCallerInfo, void* /*pvShared*/)
    {
    struct ST_CARDSETUP* pstCardSetup = static_cast < struct ST_CARDSETUP* > (pvCallerInfo);

    // ----- convert raw value to voltage -----
    short nElement = ((short*)pvSource)[offset];
    return (((float)nElement) / pstCardSetup->lMaxADCValue) * ((float)pstCardSetup->lIR) / 1000.0f;
    }
__device__ cufftCallbackLoadR d_loadCallbackPtr = CudaCallbackConvertToVolt;


__device__ void CudaCallbackStoreFFTResult (void* pvDest, size_t offset, cufftComplex element, void* pvCallerInfo, void* pvShared)
    {
    struct ST_CARDSETUP* pstCardSetup = static_cast < struct ST_CARDSETUP* > (pvCallerInfo);
    ((cufftComplex*)pvDest)[offset].x = element.x / (pstCardSetup->lLen / 2 + 1); // divide by length of signal
    ((cufftComplex*)pvDest)[offset].y = element.y / (pstCardSetup->lLen / 2 + 1); // divide by length of signal
    }
__device__ cufftCallbackStoreC d_storeCallbackPtr = CudaCallbackStoreFFTResult;

// replacement of above callback as kernel
// necessary on Windows and 32bit Linux because cufft callbacks are not available
__global__ void CudaScaleFFTResult (cufftComplex* pcompDest, const cufftComplex* pcompSource, int lLen)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    ((cufftComplex*)pcompDest)[i].x = pcompSource[i].x / (lLen / 2 + 1); // divide by length of signal
    ((cufftComplex*)pcompDest)[i].y = pcompSource[i].y / (lLen / 2 + 1); // divide by length of signal
    }


// ----- converts the complex output of a R2C FFT to a real spectrum -----
// -- pcompSource:  in  The result of a cufftExecR2C() call or similar
// -- pfDest:       out The spectrum of the FFT
__global__ void CudaKernelFFTToSpectrum (const cufftComplex* pcompSource, cufftReal* pfDest)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = sqrt (pcompSource[i].x * pcompSource[i].x + pcompSource[i].y * pcompSource[i].y); // sqrt (x^2 + y^2)
    }

// ----- converts the complex output of a R2C FFT to a real spectrum -----
// -- pcompSource:  in  The result of a cufftExecR2C() call or similar
// -- pfDest:       out The spectrum of the FFT
__global__ void CudaKernelFFTToSpectrum (const cufftComplex* pcompSource, cufftDoubleReal* pfDest)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = sqrt (pcompSource[i].x * pcompSource[i].x + pcompSource[i].y * pcompSource[i].y); // sqrt (x^2 + y^2)
    }

// ----- converts a spectrum to dBFS (dB full scale) -----
// -- pfSource: in  The spectrum of a FFT (see CudaKernelFFTToSpectrum)
// -- pfDest:   out The spectrum in dB (with respect to possible full scale)
// -- lIR_V:    in  The used input range in volts
__global__ void CudaKernelToDBFS (float* pfDest, const float* pfSource, int lIR_V)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = 20. * log10f (pfSource[i] / lIR_V);
    }
__global__ void CudaKernelToDBFS (cufftDoubleReal* pdDest, const cufftDoubleReal* pdSource, int lIR_V)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pdDest[i] = 20. * log10 (pdSource[i] / lIR_V);
    }

__global__ void CudaKernelToDBWithFactor (cufftDoubleReal* pdDest, const cufftDoubleReal* pdSource, double dFactor)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pdDest[i] = 20. * log10 (pdSource[i] / dFactor);
    }

// ----- adds samples of two arrays. Usefull for averaging. -----
// -- pfDest:       out Result. May be the same as the source values.
// -- pfSource1:    in  First summand (augend)
// -- pfSource2:    in  Second summand (addend)
__global__ void CudaKernelSum (cufftReal* pfDest, const cufftReal* pfSource1, const cufftReal* pfSource2)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = pfSource1[i] + pfSource2[i];
    }

// ----- divides by value. Usefull for averaging. -----
// -- pfDest:       out Result of division. May be the same as source.
// -- pfSource:     in  Dividend
// -- dDivideBy:    in  Divisor
__global__ void CudaKernelDivide (cufftReal* pfDest, const cufftReal* pfSource, double dDivideBy)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = pfSource[i] / dDivideBy;
    }
__global__ void CudaKernelDivide (cufftDoubleReal* pdDest, const cufftDoubleReal* pdSource, double dDivideBy)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pdDest[i] = pdSource[i] / dDivideBy;
    }

__global__ void CudaKernelMultiply (cufftReal* pfDest, const cufftReal* pfSource, double dFactor)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = pfSource[i] * dFactor;
    }
__global__ void CudaKernelMultiply (cufftDoubleReal* pdDest, const cufftDoubleReal* pdSource, double dFactor)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pdDest[i] = pdSource[i] * dFactor;
    }


// ----- Finite Impulse Response (FIR) filter. -----
// ----- Filtered signal will be shorter than the source signal by number of coefficients and will be zero-padded to match length of source signal.
// -- pfDest:           out Filtered signal
// -- pfSource:         in  Source signal that should be filtered
// -- lSourceLen:       in  Length of source signal
// -- pfCoefficients:   in  Coefficients of the FIR filter
// -- lNumCoefficients: in  Number of coefficients
__global__ void CudaKernelFIR (float* pfDest, const float* pfSource, int lSourceLen, const float* pfCoefficients, int lNumCoefficients)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    float fSum = 0.0f;

    int lDestLen = (lSourceLen - lNumCoefficients); // filtered signal is shorter than original
    if (i < lDestLen)
        {
        for (int j = 0; j < lNumCoefficients; j++)
            {
            fSum += pfCoefficients[j] * pfSource[i + lNumCoefficients - j - 1];
            }
        }

    pfDest[i] = fSum;
    }


// ----- FFT Window Functions -----

// ----- Rectangle Window. -----
// -- pfDest:       out Signal after window function has been applied
// -- pfSource:     in  Source signal
// -- dwSignalLen:  in  Length of source signal
__global__ void CudaKernelWindowRect (cufftReal* pfDest, const cufftReal* pfSource, unsigned dwSignalLen)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    // rectangular window changes nothing
    pfDest[i] = pfSource[i];
    }

// ----- Triangle Window. -----
// -- pfDest:       out Signal after window function has been applied
// -- pfSource:     in  Source signal
// -- dwSignalLen:  in  Length of source signal
__global__ void CudaKernelWindowTriangle (cufftReal* pfDest, const cufftReal* pfSource, unsigned dwSignalLen)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;

    if (i < (dwSignalLen / 2))
        pfDest[i] = pfSource[i] * 2.0 * (static_cast < double > (i) / dwSignalLen);
    else
        pfDest[i] = pfSource[i] * -2.0 * (static_cast < double > (i) / dwSignalLen) + 2.0;
    }

// ----- Sine Window. -----
// -- pfDest:       out Signal after window function has been applied
// -- pfSource:     in  Source signal
// -- dwSignalLen:  in  Length of source signal
__global__ void CudaKernelWindowSine (cufftReal* pfDest, const cufftReal* pfSource, unsigned dwSignalLen)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = pfSource[i] * static_cast < double > (pow ( sin (M_PI * static_cast < double > (i) / dwSignalLen), 3.0));
    }

// ----- Cosine Window. -----
// -- pfDest:       out Signal after window function has been applied
// -- pfSource:     in  Source signal
// -- dwSignalLen:  in  Length of source signal
__global__ void CudaKernelWindowCosine (cufftReal* pfDest, const cufftReal* pfSource, unsigned dwSignalLen)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = pfSource[i] * static_cast < double > (pow (0.5 * (1.0 -  cos (2.0 * M_PI * static_cast < double > (i) / dwSignalLen)), 2.0));
    }

// ----- Hanning Window. -----
// -- pfDest:       out Signal after window function has been applied
// -- pfSource:     in  Source signal
// -- dwSignalLen:  in  Length of source signal
__global__ void CudaKernelWindowHanning (cufftReal* pfDest, const cufftReal* pfSource, unsigned dwSignalLen)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = pfSource[i] * static_cast < double > (0.5 * (1 - cos (2.0 * M_PI * static_cast < double > (i) / dwSignalLen)));
    }

// ----- Hamming Window. -----
// -- pfDest:       out Signal after window function has been applied
// -- pfSource:     in  Source signal
// -- dwSignalLen:  in  Length of source signal
__global__ void CudaKernelWindowHamming (cufftReal* pfDest, const cufftReal* pfSource, unsigned dwSignalLen)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = pfSource[i] * static_cast < double > (0.08 + 0.46 * (1 - cos (2.0 * M_PI * static_cast < double > (i) / dwSignalLen)));
    }

// ----- Blackman Window. -----
// -- pfDest:       out Signal after window function has been applied
// -- pfSource:     in  Source signal
// -- dwSignalLen:  in  Length of source signal
__global__ void CudaKernelWindowBlackman (cufftReal* pfDest, const cufftReal* pfSource, unsigned dwSignalLen)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = pfSource[i] * static_cast < double > (0.42 - 0.50 * cos (2.0 * M_PI * static_cast < double > (i) / dwSignalLen) + 0.08 * cos (4.0 * M_PI * static_cast < double > (i) / dwSignalLen));
    }

// ----- Blackman-Harris Window. -----
// -- pfDest:       out Signal after window function has been applied
// -- pfSource:     in  Source signal
// -- dwSignalLen:  in  Length of source signal
__global__ void CudaKernelWindowBlackmanHarris (cufftReal* pfDest, const cufftReal* pfSource, unsigned dwSignalLen)
    {
    int i = blockDim.x * blockIdx.x + threadIdx.x;
    pfDest[i] = pfSource[i] *  static_cast < double > (0.42323 - 0.49755 * cos (2.0 * M_PI * static_cast < double > (i) / dwSignalLen) + 0.07922 *  cos (4.0 * M_PI * static_cast < double > (i) / dwSignalLen));
    }

#endif // SPCM_CUDA_KERNELS_H

