// ----- include standard driver header from library -----
#include "../c_header/dlltyp.h"
#include "../c_header/regs.h"
#include "../c_header/spcerr.h"
#include "../c_header/spcm_drv.h"

#include <algorithm>
#include <string>
#include <vector>
#include <iomanip>
#include <iostream>
#include <sstream>
#include <cstring> // memset

#define TIMEOUT_DISCOVERY 5000 // timeout value in ms

std::vector <std::string> vsGetAvailableVisaStrings ()
    {
    const uint32 dwMaxNumRemoteCards = 50;
    const uint32 dwMaxVisaStringLen = 50;
    const uint32 dwMaxIdnStringLen = 256;

    char* pszVisa[dwMaxNumRemoteCards] = { NULL };
    char* pszIdn[dwMaxNumRemoteCards] = { NULL };

    std::vector <std::string> vsVisa;

    // allocate memory for string list
    for (uint32 i = 0; i < dwMaxNumRemoteCards; i++)
        {
        pszVisa[i] = new char [dwMaxVisaStringLen];
        pszIdn[i] = new char [dwMaxIdnStringLen];
        memset (pszVisa[i], 0, dwMaxVisaStringLen);
        memset (pszIdn[i], 0, dwMaxIdnStringLen);
        }

    // first make discovery - check if there are any LXI compatible remote devices
    uint32 dwError = spcm_dwDiscovery ((char**)pszVisa, dwMaxNumRemoteCards, dwMaxVisaStringLen, TIMEOUT_DISCOVERY);

    // second: check from which manufacturer the devices are
    spcm_dwSendIDNRequest ((char**)pszIdn, dwMaxNumRemoteCards, dwMaxIdnStringLen);

     // ----- store VISA strings for all discovered cards and open them afterwards -----
    for (uint32 i = 0; i < dwMaxNumRemoteCards; i++)
        {
        std::string sIDN (pszIdn[i]);

        if (sIDN.size () == 0)
            break;

        if (sIDN.find("Spectrum") != std::string::npos)
            vsVisa.push_back (pszVisa[i]);
        }

    for (uint32 i = 0; i < dwMaxNumRemoteCards; i++)
        {
        delete [] pszVisa[i];
        delete [] pszIdn[i];
        }

    return vsVisa;
    }

/*
**************************************************************************
main
**************************************************************************
*/

int main ()
    {
    std::vector <std::string> vsVisaStrings;
    std::vector <int32> vlNetboxSerialnumbers;

    struct ST_REMOTE_CARD
        {
        std::string sIP;
        int32 lCardSN;
        int32 lNetboxSN;
        std::string sCard;
        std::string sVISA;
        std::string sNetbox;

        // helper function for std::sort
        bool operator<(const ST_REMOTE_CARD& rstRHS)
            {
            if (sIP < rstRHS.sIP)
                return true;
            else if (sIP == rstRHS.sIP)
                {
                if (lNetboxSN > rstRHS.lNetboxSN) // use "greater than" here to get the card with the Netbox-SN to the top
                    return true;
                else if (lNetboxSN == rstRHS.lCardSN)
                    {
                    return (lCardSN < rstRHS.lCardSN);
                    }
                }
            return false;
            }
        };
    std::vector < ST_REMOTE_CARD > vstRemoteCards;

    std::cout << "Netbox discovery running...\n\n";
    vsVisaStrings = vsGetAvailableVisaStrings ();

    for (uint32 i = 0; i < vsVisaStrings.size (); i++)
        {
        // open card
        drv_handle hCard = spcm_hOpen (vsVisaStrings[i].c_str ());
        if (!hCard)
            {
            std::cout << "no card found...\n";
            return 0;
            }

        std::string sIP (vsVisaStrings[i].substr (vsVisaStrings[i].find ("::") + 2));
        sIP = sIP.substr (0, sIP.find ("::"));

        // read some infos about the card
        int32 lCardType;
        spcm_dwGetParam_i32 (hCard, SPC_PCITYP, &lCardType);
        int32 lCardSN;
        spcm_dwGetParam_i32 (hCard, SPC_PCISERIALNO, &lCardSN);
        int32 lNetboxType;
        spcm_dwGetParam_i32 (hCard, SPC_NETBOX_TYPE, &lNetboxType);
        int32 lNetboxSN;
        spcm_dwGetParam_i32 (hCard, SPC_NETBOX_SERIALNO, &lNetboxSN);

        std::ostringstream sCard;
        switch (lCardType & TYP_SERIESMASK)
            {
            case TYP_M2ISERIES:     sCard << "M2i." << std::hex << (lCardType & 0xFFFF); break;
            case TYP_M2IEXPSERIES:  sCard << "M2i." << std::hex << (lCardType & 0xFFFF) << "-exp"; break;
            case TYP_M3ISERIES:     sCard << "M3i." << std::hex << (lCardType & 0xFFFF); break;
            case TYP_M3IEXPSERIES:  sCard << "M3i." << std::hex << (lCardType & 0xFFFF) << "-exp"; break;
            case TYP_M4IEXPSERIES:  sCard << "M4i." << std::hex << (lCardType & 0xFFFF); break;
            case TYP_M4XEXPSERIES:  sCard << "M4x." << std::hex << (lCardType & 0xFFFF); break;
            case TYP_M2PEXPSERIES:  sCard << "M2p." << std::hex << (lCardType & 0xFFFF) << "-x4"; break;
            case TYP_M5IEXPSERIES:  sCard << "M5i." << std::hex << (lCardType & 0xFFFF) << "-x16"; break;
            }

        std::ostringstream sNetbox;
        if (lNetboxType != 0)
            {
            sNetbox << "DN" << std::hex << ((lNetboxType & NETBOX_SERIES_MASK) >> 24) << "." << std::hex << ((lNetboxType & NETBOX_FAMILY_MASK) >> 16) << std::hex << ((lNetboxType & NETBOX_SPEED_MASK) >> 8) << "-" << std::setw(2) << std::setfill ('0') << std::dec << (lNetboxType & NETBOX_CHANNEL_MASK);
            }

        ST_REMOTE_CARD stRemoteCard = { sIP, lCardSN, lNetboxSN, sCard.str (), vsVisaStrings[i], sNetbox.str () };
        vstRemoteCards.push_back (stRemoteCard);

        // close card
        spcm_vClose (hCard);
        }
    std::sort (vstRemoteCards.begin (), vstRemoteCards.end ());

    if (vstRemoteCards.size ())
        std::cout << "Netboxes found:\n";
    else
        std::cout << "No Netboxes found!\n";

    std::string sLastIP;
    for (uint32 i = 0; i < vstRemoteCards.size (); i++)
        {
        if (vstRemoteCards[i].sIP != sLastIP)
            {
            std::cout << vstRemoteCards[i].sNetbox << " at " << vstRemoteCards[i].sIP << " with SN " << vstRemoteCards[i].lNetboxSN << std::endl;
            sLastIP = vstRemoteCards[i].sIP;
            }
        std::cout << "\t" << vstRemoteCards[i].sCard << " SN" << vstRemoteCards[i].lCardSN << " at " << vstRemoteCards[i].sVISA << std::endl;
        }

    return EXIT_SUCCESS;
    }


