/*
**************************************************************************

pulse_generator.cpp                                      (c) Spectrum GmbH

**************************************************************************

Example program for the pulse generator feature on M4i, M4x, M2p and M5i cards.
Different setups are shown to highlight the capabilities of the pulse
generator feature.
The setup of the card for acquisition/replay is shown in the other examples.

Documentation for the API as well as a detailed description of the hardware
can be found in the manual for each device which can be found on our website:
https://www.spectrum-instrumentation.com/en/downloads

Further information can be found online in the Knowledge Base:
https://www.spectrum-instrumentation.com/en/knowledge-base-overview

**************************************************************************
*/


// ----- include standard driver header from library -----
#include "../c_header/dlltyp.h"
#include "../c_header/regs.h"
#include "../c_header/spcerr.h"
#include "../c_header/spcm_drv.h"

#include "../common/ostools/spcm_oswrap.h"
#include "../common/ostools/spcm_ostools.h"

// ----- standard c include files -----
#include <stdio.h>
#include <string.h>
#include <stdlib.h>


// ----------------------------------------------------------------------------
// ----- this functions sets up the pulse generator for X0 to create a continuous signal
void vSingleContinuousSignal (drv_handle hCard)
    {
    int64 llPulseGenClock = 0;
    spcm_dwGetParam_i64 (hCard, SPC_XIO_PULSEGEN_CLOCK, &llPulseGenClock);

    // enable pulse generator output on XIO lines
    spcm_dwSetParam_i32 (hCard, SPCM_X0_MODE, SPCM_XMODE_PULSEGEN);

    // setup pulse generator 0 (output on X0)
    // generate a continuous signal with 1 MHz
    int32 lLenFor1MHz = static_cast < int32 > (llPulseGenClock / MEGA(1));
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_MODE,     SPCM_PULSEGEN_MODE_TRIGGERED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_LEN,      lLenFor1MHz);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_HIGH,     lLenFor1MHz / 2); // 50% HIGH, 50% LOW
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_DELAY,    0);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_LOOPS,    0); // 0: infinite
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_MUX1_SRC, SPCM_PULSEGEN_MUX1_SRC_UNUSED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_MUX2_SRC, SPCM_PULSEGEN_MUX2_SRC_SOFTWARE); // started by software force command

    // arm the trigger detection for all pulse generators
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN_ENABLE, SPCM_PULSEGEN_ENABLE0);

    // write the settings to the card
    // update the clock section to generate the programmed frequencies (SPC_SAMPLERATE)
    // and write the pulse generator settings
    spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_WRITESETUP);

    // start all pulse generators that wait for a software command
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN_COMMAND, SPCM_PULSEGEN_CMD_FORCE);
    }


// ----------------------------------------------------------------------------
// ----- this functions sets up all pulse generators to create continuous signals
// ----- with various frequencies, high/low characteristics, and offset
void vMultipleContinuousSignals (drv_handle hCard, int32 lCardType)
    {
    int64 llPulseGenClock = 0;
    spcm_dwGetParam_i64 (hCard, SPC_XIO_PULSEGEN_CLOCK, &llPulseGenClock);

    int32 lPulseGenEnableMask = 0;

    // setup pulse generator 0 (output on X0)
    // generate a continuous signal with the maximum frequency
    spcm_dwSetParam_i32 (hCard, SPCM_X0_MODE,               SPCM_XMODE_PULSEGEN); // enable pulse generator output on X0 line

    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_MODE,     SPCM_PULSEGEN_MODE_TRIGGERED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_LEN,      2);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_HIGH,     1); // 50% HIGH, 50% LOW
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_DELAY,    0);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_LOOPS,    0); // 0: infinite
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_MUX1_SRC, SPCM_PULSEGEN_MUX1_SRC_UNUSED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN0_MUX2_SRC, SPCM_PULSEGEN_MUX2_SRC_SOFTWARE); // started by software force command
    lPulseGenEnableMask |= SPCM_PULSEGEN_ENABLE0; // add pulse generator 0 to enable mask


    // setup pulse generator 1 (output on X1)
    // generate a continuous signal with ~1 MHz, 50% duty cycle
    spcm_dwSetParam_i32 (hCard, SPCM_X1_MODE,               SPCM_XMODE_PULSEGEN); // enable pulse generator output on X1 line

    int32 lLenFor1MHz = static_cast < int32 > (llPulseGenClock / MEGA(1));
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_MODE,     SPCM_PULSEGEN_MODE_TRIGGERED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_LEN,      lLenFor1MHz);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_HIGH,     lLenFor1MHz / 2); // 50% HIGH, 50% LOW
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_DELAY,    0);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_LOOPS,    0); // 0: infinite
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_MUX1_SRC, SPCM_PULSEGEN_MUX1_SRC_UNUSED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_MUX2_SRC, SPCM_PULSEGEN_MUX2_SRC_SOFTWARE); // started by software force command
    lPulseGenEnableMask |= SPCM_PULSEGEN_ENABLE1; // add pulse generator 1 to enable mask


    // setup pulse generator 2 (output on X2)
    // same signal as pulse generator 1, but with a phase shift
    spcm_dwSetParam_i32 (hCard, SPCM_X2_MODE,               SPCM_XMODE_PULSEGEN); // enable pulse generator output on X2 line

    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN2_MODE,     SPCM_PULSEGEN_MODE_TRIGGERED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN2_LEN,      lLenFor1MHz);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN2_HIGH,     lLenFor1MHz / 2); // 50% HIGH, 50% LOW
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN2_DELAY,    lLenFor1MHz / 4); // delay for 1/4 of the period to achieve a "phase shift" by 90
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN2_LOOPS,    0); // 0: infinite
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN2_MUX1_SRC, SPCM_PULSEGEN_MUX1_SRC_UNUSED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN2_MUX2_SRC, SPCM_PULSEGEN_MUX2_SRC_SOFTWARE); // started by software force command
    lPulseGenEnableMask |= SPCM_PULSEGEN_ENABLE2; // add pulse generator 2 to enable mask


    // setup pulse generator 3 (output on X3. Not available on M4i/M4x)
    if (((lCardType & TYP_SERIESMASK) != TYP_M4IEXPSERIES) && ((lCardType & TYP_SERIESMASK) != TYP_M4XEXPSERIES))
        {
        spcm_dwSetParam_i32 (hCard, SPCM_X3_MODE, SPCM_XMODE_PULSEGEN); // enable pulse generator output on X3 lines

        // generate a continuous signal with ~500 kHz after the first edge on pulse generator 2 occurred, and delay the start for two periods of the 1MHz signal.
        int32 lLenFor500kHz = static_cast < int32 > (llPulseGenClock / KILO(500));
        spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN3_MODE,     SPCM_PULSEGEN_MODE_TRIGGERED);
        spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN3_LEN,      lLenFor500kHz);
        spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN3_HIGH,     9*lLenFor500kHz / 10); // 90% HIGH, 10% LOW
        spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN3_DELAY,    2*lLenFor1MHz); // delay for two periods of 1MHz signal
        spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN3_LOOPS,    0); // 0: infinite
        spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN3_MUX1_SRC, SPCM_PULSEGEN_MUX1_SRC_UNUSED);
        spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN3_MUX2_SRC, SPCM_PULSEGEN_MUX2_SRC_PULSEGEN2); // started by first edge of pulse generator 2
        lPulseGenEnableMask |= SPCM_PULSEGEN_ENABLE3; // add pulse generator 3 to enable mask
        }

    // arm the trigger detection for all pulse generators
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN_ENABLE, lPulseGenEnableMask);

    // write the settings to the card
    // update the clock section to generate the programmed frequencies (SPC_SAMPLERATE)
    // and write the pulse generator settings
    spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_WRITESETUP);

    // start all pulse generators that wait for a software command
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN_COMMAND, SPCM_PULSEGEN_CMD_FORCE);
    }


// ----------------------------------------------------------------------------
// ----- This functions creates a single pulse on X1 after the card has detected a trigger event.
// ----- Possible application: "change" length of trigger output signal, or create a delayed version of trigger output
void vCreateDelayedTriggerPulse (drv_handle hCard)
    {
    // we use software trigger to trigger the card immediately after it has been started (start command at end of this function)
    spcm_dwSetParam_i64 (hCard, SPC_TRIG_ORMASK, SPC_TMASK_SOFTWARE);

    int64 llPulseGenClock = 0;
    spcm_dwGetParam_i64 (hCard, SPC_XIO_PULSEGEN_CLOCK, &llPulseGenClock);

    spcm_dwSetParam_i32 (hCard, SPCM_X0_MODE, SPCM_XMODE_TRIGOUT);  // trigger output for reference on X0
    spcm_dwSetParam_i32 (hCard, SPCM_X1_MODE, SPCM_XMODE_PULSEGEN); // enable pulse generator output on X1

    int32 lLen_1ms    = static_cast < int32 > (llPulseGenClock * 0.001 + 1); // +1 because the HIGH area needs to be at least one sample less than length, so we increase length by one to get the calculated HIGH time
    int32 lOffset_5us = static_cast < int32 > (llPulseGenClock * 0.000005);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_MODE,     SPCM_PULSEGEN_MODE_SINGLESHOT);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_LEN,      lLen_1ms);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_HIGH,     lLen_1ms - 1);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_DELAY,    lOffset_5us); // delay the pulse for 5us
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_LOOPS,    1); // just once
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_MUX1_SRC, SPCM_PULSEGEN_MUX1_SRC_UNUSED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_MUX2_SRC, SPCM_PULSEGEN_MUX2_SRC_CARDTRIGGER); // started by card's trigger event

    // arm the trigger detection for our selected pulse generator
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN_ENABLE, SPCM_PULSEGEN_ENABLE1);

    // write the settings to the card
    // update the clock section to generate the programmed frequencies (SPC_SAMPLERATE)
    // and write the pulse generator settings
    spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_WRITESETUP);

    // start card. Its trigger event will start the pulse generator
    spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_START | M2CMD_CARD_ENABLETRIGGER);
    }


// ----------------------------------------------------------------------------
// ----- This function takes a trigger signal on X2 and gives a pulse that is 
// ----- synchronized to the card's clock back out on X1. This pulse can then
// ----- be used to trigger the card and other external equipment without the
// ----- normal 1 clock jitter for external trigger events.
void vSynchronizeExternalPulse (drv_handle hCard)
    {
    int64 llPulseGenClock = 0;
    spcm_dwGetParam_i64 (hCard, SPC_XIO_PULSEGEN_CLOCK, &llPulseGenClock);

    spcm_dwSetParam_i32 (hCard, SPCM_X2_MODE, SPCM_XMODE_ASYNCIN);  // we only need to set X2 to some kind of input, and ASYNCIN is available on all card series
    spcm_dwSetParam_i32 (hCard, SPCM_X1_MODE, SPCM_XMODE_PULSEGEN); // enable pulse generator output on X1

    int32 lLen_1ms    = static_cast < int32 > (llPulseGenClock * 0.001 + 1); // +1 because the HIGH area needs to be at least one sample less than length, so we increase length by one to get the calculated HIGH time
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_MODE,     SPCM_PULSEGEN_MODE_TRIGGERED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_LEN,      lLen_1ms);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_HIGH,     lLen_1ms - 1);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_DELAY,    0);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_LOOPS,    1); // just once per event
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_MUX1_SRC, SPCM_PULSEGEN_MUX1_SRC_UNUSED);
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN1_MUX2_SRC, SPCM_PULSEGEN_MUX2_SRC_XIO2); // started by rising edge on X2

    // arm the trigger detection for our selected pulse generator
    spcm_dwSetParam_i32 (hCard, SPC_XIO_PULSEGEN_ENABLE, SPCM_PULSEGEN_ENABLE1);

    // write the settings to the card
    spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_WRITESETUP);
    }


/*
**************************************************************************
main
**************************************************************************
*/

int main ()
    {
    drv_handle  hCard;
    int32       lCardType, lSerialNumber;
    char        szName[50];

    // ------------------------------------------------------------------------
    // we try to open one card and printout some information on it

    // uncomment the second line and replace the IP address to use remote
    // cards like in a digitizerNETBOX
    sprintf (szName, "/dev/spcm0");
    // sprintf (szName, "TCPIP::192.168.1.10::inst0::INSTR");
    hCard = spcm_hOpen (szName);

    // not one card found
    if (!hCard)
        {
        printf ("no card found...\n");
        return 0;
        }

    // read card type name
    char acCardType[20] = {};
    spcm_dwGetParam_ptr (hCard, SPC_PCITYP, acCardType, sizeof (acCardType));

    // read out some info and print it
    spcm_dwGetParam_i32 (hCard, SPC_PCITYP,            &lCardType);
    spcm_dwGetParam_i32 (hCard, SPC_PCISERIALNO,       &lSerialNumber);
    printf ("Found: %s sn %05d\n", acCardType, lSerialNumber);

    // pulse generators are not available on M2i and M3i cards
    switch (lCardType & TYP_SERIESMASK)
        {
        case TYP_M2ISERIES:
        case TYP_M2IEXPSERIES:
        case TYP_M3ISERIES:
        case TYP_M3IEXPSERIES:
            {
            printf ("ERROR: pulse generators not available on this card\n");
            return EXIT_FAILURE;
            }
        }

    // on newer cards the availability depends on a (paid) feature
    // so we check if this card has the pulse generator feature installed
    int32 lFeatures = 0;
    spcm_dwGetParam_i32 (hCard, SPC_PCIEXTFEATURES, &lFeatures);
    if ((lFeatures & SPCM_FEAT_EXTFW_PULSEGEN) == 0)
        {
        printf ("ERROR: pulse generators not supported by this card\n");
        return EXIT_FAILURE;
        }

    // first we set up the channel selection and the clock
    // for this example we enable only one channel to be able to use max sampling rate on all card types
    // ! changing the card settings while pulse generators are active causes a stop and restart of the pulse generators
    spcm_dwSetParam_i32 (hCard, SPC_CHENABLE, CHANNEL0);
    int64 llMaxSR = 0;
    spcm_dwGetParam_i64 (hCard, SPC_PCISAMPLERATE, &llMaxSR);
    spcm_dwSetParam_i64 (hCard, SPC_SAMPLERATE, llMaxSR);

    // start the pulse generator
    // uncomment one of the following function calls. Each will set up the pulse generators differently.
    vSingleContinuousSignal (hCard);
    //vMultipleContinuousSignals (hCard, lCardType);
    //vCreateDelayedTriggerPulse (hCard);
    //vSynchronizeExternalPulse (hCard);

    // wait until user presses a key
    printf ("\nPress a key to stop the pulse generator(s) ");
    cGetch (); 

    // stop the pulse generators
    spcm_dwSetParam_i32(hCard, SPC_XIO_PULSEGEN_ENABLE, 0);
    spcm_dwSetParam_i32(hCard, SPC_M2CMD, M2CMD_CARD_WRITESETUP);

    spcm_vClose (hCard);

    return EXIT_SUCCESS;
    }

