/*
**************************************************************************

dds_starhub_sync.cpp                                     (c) Spectrum GmbH

**************************************************************************

Example for SpcMDrv based analog generation cards with DDS feature installed.

Information about the different products and their drivers can be found
online in the Knowledge Base:
https://www.spectrum-instrumentation.com/en/platform-driver-and-series-differences

Shows different ways to use the DDS feature to generate single frequencies,
multiple frequencies and AM or FM modulation.

Feel free to use this source for own projects and modify it in any kind

Documentation for the API as well as a detailed description of the hardware
can be found in the manual for each device which can be found on our website:
https://www.spectrum-instrumentation.com/en/downloads

Further information can be found online in the Knowledge Base:
https://www.spectrum-instrumentation.com/en/knowledge-base-overview

**************************************************************************
*/


// ----- include standard driver header from library -----
#include "../../c_header/dlltyp.h"
#include "../../c_header/regs.h"
#include "../../c_header/spcerr.h"
#include "../../c_header/spcm_drv.h"

#include "../../common/ostools/spcm_oswrap.h"
#include "../../common/ostools/spcm_ostools.h"

// ----- standard c include files -----
#include <cstdio>
#include <cstdlib>


/*
**************************************************************************
main
**************************************************************************
*/


int main()
    {
	drv_handle  ahCard[2];
	drv_handle  hSync = NULL_HANDLE;
    bool        bError = false;
    char        szErrorTextBuffer[ERRORTEXTLEN], szName[50];
	long        lCardCount = 2;
	uint32      dwErr;

    // ------------------------------------------------------------------------
    // uncomment the second line and replace the IP address to use remote
    // cards like in a generatorNETBOX
    int32 lCardIdx = 0;

	for (lCardIdx = 0; lCardIdx < lCardCount; lCardIdx++)
		{
		sprintf(szName, "/dev/spcm%d", lCardIdx);
		// sprintf (szName, "TCPIP::192.168.1.10::inst%d::INSTR", lCardIdx);
		ahCard[lCardIdx] = spcm_hOpen(szName);

		// no card found
		if (!ahCard[lCardIdx])
			{
			printf("no card found...\n");
			return 0;
			}

		// read card type name
		char acCardType[20] = {};
		spcm_dwGetParam_ptr(ahCard[lCardIdx], SPC_PCITYP, acCardType, sizeof(acCardType));

		// read out some info and print it
		int32 lCardType = 0;
		spcm_dwGetParam_i32(ahCard[lCardIdx], SPC_PCITYP, &lCardType);
		int32 lSerialNumber = 0;
		spcm_dwGetParam_i32(ahCard[lCardIdx], SPC_PCISERIALNO, &lSerialNumber);
		int32 lFncType = 0;
		spcm_dwGetParam_i32(ahCard[lCardIdx], SPC_FNCTYPE, &lFncType);

		switch (lFncType)
			{
			// only D/A cards allowed in example
			case SPCM_TYPE_AO:
				printf("Found: %s sn %05d\n", acCardType, lSerialNumber);
				break;

				// print error for all other types
			default:
				printf("Card: %s sn %05d not supported by example\n", acCardType, lSerialNumber);
				spcm_vClose(ahCard[lCardIdx]);
				return EXIT_FAILURE;
			}

		// ----- check if DDS feature is available -----
		int32 lFWFeatures = 0;
		spcm_dwGetParam_i32(ahCard[lCardIdx], SPC_PCIEXTFEATURES, &lFWFeatures);
		if ((lFWFeatures & SPCM_FEAT_EXTFW_DDS) == 0)
			{
			printf("ERROR: this example requires the DDS feature to be installed on the card.\n");
			spcm_vClose(ahCard[lCardIdx]);
			return EXIT_FAILURE;
			}

		// ----- calculate available number of channels on this card -----
		int32 lNumModules = 0;
		spcm_dwGetParam_i32(ahCard[lCardIdx], SPC_MIINST_MODULES, &lNumModules);
		int32 lNumChPerModule = 0;
		spcm_dwGetParam_i32(ahCard[lCardIdx], SPC_MIINST_CHPERMODULE, &lNumChPerModule);
		int32 lNumCh = lNumModules * lNumChPerModule;


		// ------------------------------------------------------------------------
		// do a simple standard setup for the card
		spcm_dwSetParam_i32(ahCard[lCardIdx], SPC_CHENABLE, (0x1 << lNumCh) - 1);  // enable all channels
		spcm_dwSetParam_i32(ahCard[lCardIdx], SPC_CARDMODE, SPC_REP_STD_DDS);      // DDS mode
		spcm_dwSetParam_i32(ahCard[lCardIdx], SPC_CLOCKMODE, SPC_CM_INTPLL);        // clock mode internal PLL

		// setup the channels
		for (int lChIdx = 0; lChIdx < lNumCh; ++lChIdx)
			{
			spcm_dwSetParam_i32(ahCard[lCardIdx], SPC_AMP0 + 100 * lChIdx, 1000);      // +-1V output amplitude
			spcm_dwSetParam_i32(ahCard[lCardIdx], SPC_ENABLEOUT0 + 100 * lChIdx, 1);   // enable output of channel
			}

		// ----- write the non-DDS setup to the card and check for error -----
		if (spcm_dwSetParam_i32(ahCard[lCardIdx], SPC_M2CMD, M2CMD_CARD_WRITESETUP) != ERR_OK)
			{
			spcm_dwGetErrorInfo_i32(ahCard[lCardIdx], NULL, NULL, szErrorTextBuffer);
			printf("ERROR: %s\n", szErrorTextBuffer);
			printf("Press Enter to Exit");
			getchar();

			spcm_vClose(ahCard[lCardIdx]);

			return EXIT_FAILURE;
			}
		}

	// ------------------------------------------------------------------------
	// try to open the star-hub
	if (!bError)
		{
		hSync = spcm_hOpen("sync0");
		if (!hSync)
			{
			printf("no star-hub found. This is essential for the example ...\n");
			spcm_vClose(ahCard[0]);
			spcm_vClose(ahCard[1]);
			return EXIT_FAILURE;
			}
		else
			printf("Found Star-Hub ...\n");
		}

    // ------------------------------------------------------------------------
    // setup the DDS
    // first reset it
	spcm_dwSetParam_i32 (ahCard[0], SPC_DDS_CMD, SPCM_DDS_CMD_RESET);
	spcm_dwSetParam_i32 (ahCard[1], SPC_DDS_CMD, SPCM_DDS_CMD_RESET);

    // then program 

	spcm_dwSetParam_d64(ahCard[0], SPC_DDS_CORE0_AMP, 0.4);
	spcm_dwSetParam_d64(ahCard[0], SPC_DDS_CORE0_FREQ, KILO(10));
	spcm_dwSetParam_i32(ahCard[0], SPC_DDS_CMD, SPCM_DDS_CMD_EXEC_AT_TRG);
	spcm_dwSetParam_i32(ahCard[0], SPC_DDS_CMD, SPCM_DDS_CMD_WRITE_TO_CARD);

	spcm_dwSetParam_d64(ahCard[1], SPC_DDS_CORE0_AMP, 0.4);
	spcm_dwSetParam_d64(ahCard[1], SPC_DDS_CORE0_FREQ, KILO(10));
	spcm_dwSetParam_i32(ahCard[1], SPC_DDS_CMD, SPCM_DDS_CMD_EXEC_AT_TRG);
	spcm_dwSetParam_i32(ahCard[1], SPC_DDS_CMD, SPCM_DDS_CMD_WRITE_TO_CARD);

    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////

	spcm_dwSetParam_i32(hSync, SPC_SYNC_ENABLEMASK, (1 << lCardCount) - 1);

    // Trigger has to be assigned after set of SPC_SYNC_ENABLEMASK!
    spcm_dwSetParam_i32(ahCard[0], SPC_TRIG_ORMASK, SPC_TMASK_EXT0);
    spcm_dwSetParam_i32(ahCard[0], SPC_TRIG_ANDMASK, 0);

    // The other Card is automatically triggered by the other!
    spcm_dwSetParam_i32(ahCard[1], SPC_TRIG_ORMASK, 0);     
    spcm_dwSetParam_i32(ahCard[1], SPC_TRIG_ANDMASK, 0);



    // Start command including enable of trigger engine -> error check
    if (spcm_dwSetParam_i32 (hSync, SPC_M2CMD, M2CMD_CARD_START | M2CMD_CARD_ENABLETRIGGER) != ERR_OK)
        {
        spcm_dwGetErrorInfo_i32 (hSync, NULL, NULL, szErrorTextBuffer);
        printf ("ERROR: %s\n", szErrorTextBuffer);
        }

    spcm_dwSetParam_i32(ahCard[0], SPC_M2CMD, M2CMD_CARD_FORCETRIGGER);
    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////

    printf ("Press Enter to Exit");
    getchar ();

	spcm_dwSetParam_i32(hSync, SPC_M2CMD, M2CMD_CARD_STOP);
	spcm_vClose (ahCard[0]);
	spcm_vClose (ahCard[1]);

    return EXIT_SUCCESS;
    }
