/*
**************************************************************************

dds_complex_ramps.cpp                                    (c) Spectrum GmbH

**************************************************************************

Example for SpcMDrv based analog generation cards with DDS feature installed.

Information about the different products and their drivers can be found
online in the Knowledge Base:
https://www.spectrum-instrumentation.com/en/platform-driver-and-series-differences

Shows how to generate s-shaped-ramps in frequency

Feel free to use this source for own projects and modify it in any kind

Documentation for the API as well as a detailed description of the hardware
can be found in the manual for each device which can be found on our website:
https://www.spectrum-instrumentation.com/en/downloads

Further information can be found online in the Knowledge Base:
https://www.spectrum-instrumentation.com/en/knowledge-base-overview

**************************************************************************
*/


// ----- include standard driver header from library -----
#include "../../c_header/dlltyp.h"
#include "../../c_header/regs.h"
#include "../../c_header/spcerr.h"
#include "../../c_header/spcm_drv.h"

#include "../../common/ostools/spcm_oswrap.h"
#include "../../common/ostools/spcm_ostools.h"

#define _USE_MATH_DEFINES // for M_PI
#include <cmath>

// ----- standard c include files -----
#include <cstdio>
#include <cstdlib>
#include <vector>


/*
**************************************************************************
main
**************************************************************************
*/


int main()
    {
    drv_handle  hCard = NULL_HANDLE;
    bool        bError = false;
    char        szErrorTextBuffer[ERRORTEXTLEN], szName[50];

    // ------------------------------------------------------------------------
    // uncomment the second line and replace the IP address to use remote
    // cards like in a generatorNETBOX
    int32 lCardIdx = 0;
    sprintf(szName, "/dev/spcm%d", lCardIdx);
    // sprintf (szName, "TCPIP::192.168.1.10::inst%d::INSTR", lCardIdx);
    hCard = spcm_hOpen (szName);

    // no card found
    if (!hCard)
        {
        printf ("no card found...\n");
        return 0;
        }

	// read card type name
	char acCardType[20] = {};
	spcm_dwGetParam_ptr(hCard, SPC_PCITYP, acCardType, sizeof(acCardType));

    // read out some info and print it
    int32 lCardType = 0;
    spcm_dwGetParam_i32 (hCard, SPC_PCITYP, &lCardType);
    int32 lSerialNumber = 0;
    spcm_dwGetParam_i32 (hCard, SPC_PCISERIALNO, &lSerialNumber);
    int32 lFncType = 0;
    spcm_dwGetParam_i32 (hCard, SPC_FNCTYPE, &lFncType);

    switch (lFncType)
        {
        // only D/A cards allowed in example
        case SPCM_TYPE_AO:
            printf ("Found: %s sn %05d\n", acCardType, lSerialNumber);
            break;

        // print error for all other types
        default:
            printf ("Card: %s sn %05d not supported by example\n", acCardType, lSerialNumber);
            spcm_vClose (hCard);
            return EXIT_FAILURE;
        }

    // ----- check if DDS feature is available -----
    int32 lFWFeatures = 0;
    spcm_dwGetParam_i32 (hCard, SPC_PCIEXTFEATURES, &lFWFeatures);
    if ((lFWFeatures & SPCM_FEAT_EXTFW_DDS) == 0)
        {
        printf ("ERROR: this example requires the DDS feature to be installed on the card.\n");
        spcm_vClose(hCard);
        return EXIT_FAILURE;
        }

    // ----- calculate available number of channels on this card -----
    int32 lNumModules = 0;
    spcm_dwGetParam_i32 (hCard, SPC_MIINST_MODULES, &lNumModules);
    int32 lNumChPerModule = 0;
    spcm_dwGetParam_i32 (hCard, SPC_MIINST_CHPERMODULE, &lNumChPerModule);
    int32 lNumCh = lNumModules * lNumChPerModule;


    // ------------------------------------------------------------------------
    // do a simple standard setup for the card
    spcm_dwSetParam_i32 (hCard, SPC_CHENABLE,     (0x1 << lNumCh) - 1);  // enable all channels
    spcm_dwSetParam_i32 (hCard, SPC_CARDMODE,     SPC_REP_STD_DDS);      // DDS mode
    spcm_dwSetParam_i32 (hCard, SPC_CLOCKMODE,    SPC_CM_INTPLL);        // clock mode internal PLL
    spcm_dwSetParam_i32 (hCard, SPC_TRIG_ORMASK,  SPC_TMASK_NONE);       // Defines Card Trigger Used by DDS, None => Only Force Trigger is enabled

    // setup the channels
    for (int lChIdx = 0; lChIdx < lNumCh; ++lChIdx)
        {
        spcm_dwSetParam_i32 (hCard, SPC_AMP0 + 100 * lChIdx, 1000);      // +-1V output amplitude
        spcm_dwSetParam_i32 (hCard, SPC_ENABLEOUT0 + 100 * lChIdx, 1);   // enable output of channel
        }

    // ----- write the non-DDS setup to the card and check for error -----
    if (spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_WRITESETUP) != ERR_OK)
        {
        spcm_dwGetErrorInfo_i32 (hCard, NULL, NULL, szErrorTextBuffer);
        printf ("ERROR: %s\n", szErrorTextBuffer);
        printf ("Press Enter to Exit");
        getchar ();

        spcm_vClose(hCard);

        return EXIT_FAILURE;
        }

    // ------------------------------------------------------------------------
    // setup the DDS

    // first reset it
    spcm_dwSetParam_i32 (hCard, SPC_DDS_CMD, SPCM_DDS_CMD_RESET);

	int32 lNumCores = 0;
	spcm_dwGetParam_i32(hCard, SPC_DDS_NUM_CORES, &lNumCores);

	// 20 Carriers from 90 to 110 MHz
	double dFirstInitFreq_Hz = MEGA(90);
	double dDeltaInitFreq_Hz = MEGA(1);
	// to 20 Carriers from 95 to 105 MHz
	double dFirstFinalFreq_Hz = MEGA(95);
	double dDeltaFinalFreq_Hz = KILO(500);

	// Ramp settings
	int32 lNumSegments = 16;
	double dTotalTime_s = 5.0;
	double dDeltaTime_s = dTotalTime_s / lNumSegments;

	// Calculate the slopes
	std::vector<std::vector<double>> vvdFreqSlopeList;
	double d_dy = 0;
	double dSlope = 0;
	std::vector<double> vdFreqSlopeList;
	for (unsigned int dwSegment = 0; dwSegment < lNumSegments; dwSegment++)
		{
		for (unsigned int dwCore = 0; dwCore < lNumCores; dwCore++) {
			d_dy = ((dFirstFinalFreq_Hz + dwCore * dDeltaFinalFreq_Hz) - (dFirstInitFreq_Hz + dwCore * dDeltaInitFreq_Hz)) * (0.5*cos(M_PI*dwSegment / lNumSegments) - 0.5*cos(M_PI*(dwSegment+1) / lNumSegments));
			dSlope = d_dy / dDeltaTime_s;
			vdFreqSlopeList.push_back(dSlope);
			}
		vvdFreqSlopeList.push_back(vdFreqSlopeList);
		vdFreqSlopeList.clear();
		}

    // then program 
	spcm_dwSetParam_i32(hCard, SPC_DDS_TRG_SRC, SPCM_DDS_TRG_SRC_TIMER);
	spcm_dwSetParam_d64(hCard, SPC_DDS_TRG_TIMER, 2.0);
	// Create the carrier
	for (unsigned int dwCounter = 0; dwCounter < (unsigned int)lNumCores; dwCounter++) {
		spcm_dwSetParam_d64(hCard, SPC_DDS_CORE0_AMP + dwCounter, 0.45 / lNumCores);
		spcm_dwSetParam_d64(hCard, SPC_DDS_CORE0_FREQ + dwCounter, dFirstInitFreq_Hz + dwCounter*dDeltaInitFreq_Hz);
		}
	spcm_dwSetParam_i32(hCard, SPC_DDS_CMD, SPCM_DDS_CMD_EXEC_AT_TRG);
	spcm_dwSetParam_i32(hCard, SPC_DDS_CMD, SPCM_DDS_CMD_WRITE_TO_CARD);

	// STEP 1 - Start the ramp
	double dFrequencySlope;
	spcm_dwSetParam_d64(hCard, SPC_DDS_TRG_TIMER, dDeltaTime_s);
	for (unsigned int dwSegment = 0; dwSegment < lNumSegments; dwSegment++)
		{
		for (unsigned int dwCore = 0; dwCore < lNumCores; dwCore++) {
			dFrequencySlope = vvdFreqSlopeList[dwSegment][dwCore];
			spcm_dwSetParam_d64(hCard, SPC_DDS_CORE0_FREQ_SLOPE + dwCore, dFrequencySlope);
			}
		spcm_dwSetParam_i32(hCard, SPC_DDS_CMD, SPCM_DDS_CMD_EXEC_AT_TRG);
		}
	spcm_dwSetParam_i32(hCard, SPC_DDS_CMD, SPCM_DDS_CMD_WRITE_TO_CARD);

	// STEP 2 - Stop the ramp
	for (unsigned int dwCore = 0; dwCore < lNumCores; dwCore++) {
		spcm_dwSetParam_d64(hCard, SPC_DDS_CORE0_FREQ_SLOPE + dwCore, 0);
		spcm_dwSetParam_d64(hCard, SPC_DDS_CORE0_FREQ + dwCore, dFirstFinalFreq_Hz + dDeltaFinalFreq_Hz * dwCore);
		}
	spcm_dwSetParam_i32(hCard, SPC_DDS_CMD, SPCM_DDS_CMD_EXEC_AT_TRG);
	spcm_dwSetParam_i32(hCard, SPC_DDS_CMD, SPCM_DDS_CMD_WRITE_TO_CARD);

    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////

    // start command including enable of trigger engine -> error check
    if (spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_START | M2CMD_CARD_ENABLETRIGGER) != ERR_OK)
        {
        spcm_dwGetErrorInfo_i32 (hCard, NULL, NULL, szErrorTextBuffer);
        printf ("ERROR: %s\n", szErrorTextBuffer);
        }
	spcm_dwSetParam_i32(hCard, SPC_M2CMD, M2CMD_CARD_FORCETRIGGER);

    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////
    
    printf ("Press Enter to Exit");
    getchar ();

	spcm_dwSetParam_i32(hCard, SPC_M2CMD, M2CMD_CARD_STOP);
    spcm_vClose (hCard);

    return EXIT_SUCCESS;
    }
