/*
**************************************************************************

dds_multiple_carriers_channels.cpp                       (c) Spectrum GmbH

**************************************************************************

Example for SpcMDrv based analog generation cards with DDS feature installed.

Information about the different products and their drivers can be found
online in the Knowledge Base:
https://www.spectrum-instrumentation.com/en/platform-driver-and-series-differences

Shows different ways to use the DDS feature to generate single frequencies,
multiple frequencies and AM or FM modulation.

Feel free to use this source for own projects and modify it in any kind

Documentation for the API as well as a detailed description of the hardware
can be found in the manual for each device which can be found on our website:
https://www.spectrum-instrumentation.com/en/downloads

Further information can be found online in the Knowledge Base:
https://www.spectrum-instrumentation.com/en/knowledge-base-overview

**************************************************************************
*/


// ----- include standard driver header from library -----
#include "../../c_header/dlltyp.h"
#include "../../c_header/regs.h"
#include "../../c_header/spcerr.h"
#include "../../c_header/spcm_drv.h"

#include "../../common/ostools/spcm_oswrap.h"
#include "../../common/ostools/spcm_ostools.h"

// ----- standard c include files -----
#include <cstdio>
#include <cstdlib>


/*
**************************************************************************
main
**************************************************************************
*/


int main()
    {
    drv_handle  hCard = NULL_HANDLE;
    bool        bError = false;
    char        szErrorTextBuffer[ERRORTEXTLEN], szName[50];

    // ------------------------------------------------------------------------
    // uncomment the second line and replace the IP address to use remote
    // cards like in a generatorNETBOX
    int32 lCardIdx = 0;
    sprintf(szName, "/dev/spcm%d", lCardIdx);
    // sprintf (szName, "TCPIP::192.168.1.10::inst%d::INSTR", lCardIdx);
    hCard = spcm_hOpen (szName);

    // no card found
    if (!hCard)
        {
        printf ("no card found...\n");
        return 0;
        }

	// read card type name
	char acCardType[20] = {};
	spcm_dwGetParam_ptr(hCard, SPC_PCITYP, acCardType, sizeof(acCardType));

    // read out some info and print it
    int32 lCardType = 0;
    spcm_dwGetParam_i32 (hCard, SPC_PCITYP, &lCardType);
    int32 lSerialNumber = 0;
    spcm_dwGetParam_i32 (hCard, SPC_PCISERIALNO, &lSerialNumber);
    int32 lFncType = 0;
    spcm_dwGetParam_i32 (hCard, SPC_FNCTYPE, &lFncType);

    switch (lFncType)
        {
        // only D/A cards allowed in example
        case SPCM_TYPE_AO:
            printf ("Found: %s sn %05d\n", acCardType, lSerialNumber);
            break;

        // print error for all other types
        default:
            printf ("Card: %s sn %05d not supported by example\n", acCardType, lSerialNumber);
            spcm_vClose (hCard);
            return EXIT_FAILURE;
        }

    // ----- check if DDS feature is available -----
    int32 lFWFeatures = 0;
    spcm_dwGetParam_i32 (hCard, SPC_PCIEXTFEATURES, &lFWFeatures);
    if ((lFWFeatures & SPCM_FEAT_EXTFW_DDS) == 0)
        {
        printf ("ERROR: this example requires the DDS feature to be installed on the card.\n");
        spcm_vClose(hCard);
        return EXIT_FAILURE;
        }

    // ----- calculate available number of channels on this card -----
    int32 lNumModules = 0;
    spcm_dwGetParam_i32 (hCard, SPC_MIINST_MODULES, &lNumModules);
    int32 lNumChPerModule = 0;
    spcm_dwGetParam_i32 (hCard, SPC_MIINST_CHPERMODULE, &lNumChPerModule);
    int32 lNumCh = lNumModules * lNumChPerModule;


    // ------------------------------------------------------------------------
    // do a simple standard setup for the card
    spcm_dwSetParam_i32 (hCard, SPC_CHENABLE,     (0x1 << lNumCh) - 1);  // enable all channels
    spcm_dwSetParam_i32 (hCard, SPC_CARDMODE,     SPC_REP_STD_DDS);      // DDS mode
    spcm_dwSetParam_i32 (hCard, SPC_CLOCKMODE,    SPC_CM_INTPLL);        // clock mode internal PLL
    spcm_dwSetParam_i32 (hCard, SPC_TRIG_ORMASK,  SPC_TMASK_NONE);       // Defines Card Trigger Used by DDS, None => Only Force Trigger is enabled

    // setup the channels
    for (int lChIdx = 0; lChIdx < lNumCh; ++lChIdx)
        {
        spcm_dwSetParam_i32 (hCard, SPC_AMP0 + 100 * lChIdx, 1000);      // +-1V output amplitude
        spcm_dwSetParam_i32 (hCard, SPC_ENABLEOUT0 + 100 * lChIdx, 1);   // enable output of channel
        }

    // ----- write the non-DDS setup to the card and check for error -----
    if (spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_WRITESETUP) != ERR_OK)
        {
        spcm_dwGetErrorInfo_i32 (hCard, NULL, NULL, szErrorTextBuffer);
        printf ("ERROR: %s\n", szErrorTextBuffer);
        printf ("Press Enter to Exit");
        getchar ();

        spcm_vClose(hCard);

        return EXIT_FAILURE;
        }

    // ------------------------------------------------------------------------
    // setup the DDS
    // first reset it
    spcm_dwSetParam_i32 (hCard, SPC_DDS_CMD, SPCM_DDS_CMD_RESET);

    // then the program 
	if (lNumCh == 2) {
		spcm_dwSetParam_i32(hCard, SPC_DDS_CORES_ON_CH1, SPCM_DDS_CORE8 | SPCM_DDS_CORE9 | SPCM_DDS_CORE10 | SPCM_DDS_CORE11 | SPCM_DDS_CORE20);
		}
	else if (lNumCh == 4) {
		spcm_dwSetParam_i32(hCard, SPC_DDS_CORES_ON_CH1, SPCM_DDS_CORE8 | SPCM_DDS_CORE9 | SPCM_DDS_CORE10 | SPCM_DDS_CORE11 | SPCM_DDS_CORE20);
		spcm_dwSetParam_i32(hCard, SPC_DDS_CORES_ON_CH2, SPCM_DDS_CORE12 | SPCM_DDS_CORE13 | SPCM_DDS_CORE14 | SPCM_DDS_CORE15 | SPCM_DDS_CORE21);
		spcm_dwSetParam_i32(hCard, SPC_DDS_CORES_ON_CH3, SPCM_DDS_CORE16 | SPCM_DDS_CORE17 | SPCM_DDS_CORE18 | SPCM_DDS_CORE19 | SPCM_DDS_CORE22);
		}
	// spcm_dwSetParam_i32(hCard, SPC_DDS_CONNECTIONS, SPCM_DDS_CONN_CORE8_11_TO_CH1 | SPCM_DDS_CONN_CORE12_15_TO_CH2 | SPCM_DDS_CONN_CORE16_19_TO_CH3);

	int32 lNumCores = 0;
	spcm_dwGetParam_i32(hCard, SPC_DDS_NUM_CORES, &lNumCores);

	printf("lNumCores: %d\n", lNumCores);

	double dStartFreq_Hz = MEGA(5);
	double dDeltaFreq_Hz = KILO(500);

	for (int lCore = 0; lCore < lNumCores; lCore++)
		{
		spcm_dwSetParam_d64(hCard, SPC_DDS_CORE0_AMP + lCore, 0.4 / lNumCores);
		spcm_dwSetParam_d64(hCard, SPC_DDS_CORE0_FREQ + lCore, dStartFreq_Hz + lCore * dDeltaFreq_Hz);
		}
	spcm_dwSetParam_i32(hCard, SPC_DDS_CMD, SPCM_DDS_CMD_EXEC_AT_TRG);
	spcm_dwSetParam_i32(hCard, SPC_DDS_CMD, SPCM_DDS_CMD_WRITE_TO_CARD);

    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////

    // start command including enable of trigger engine -> error check
    if (spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_START | M2CMD_CARD_ENABLETRIGGER) != ERR_OK)
        {
        spcm_dwGetErrorInfo_i32 (hCard, NULL, NULL, szErrorTextBuffer);
        printf ("ERROR: %s\n", szErrorTextBuffer);
        }

    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////
    spcm_dwSetParam_i32(hCard, SPC_M2CMD, M2CMD_CARD_FORCETRIGGER);

    printf ("Press Enter to Exit");
    getchar ();

	spcm_dwSetParam_i32(hCard, SPC_M2CMD, M2CMD_CARD_STOP);
    spcm_vClose (hCard);

    return EXIT_SUCCESS;
    }
