/*
**************************************************************************

simple_rep_std.cpp                                       (c) Spectrum GmbH

**************************************************************************

Example for all SpcMDrv based analog generation cards.

Information about the different products and their drivers can be found
online in the Knowledge Base:
https://www.spectrum-instrumentation.com/en/platform-driver-and-series-differences

Shows a standard mode example using only the few necessary commands

Feel free to use this source for own projects and modify it in any kind

Documentation for the API as well as a detailed description of the hardware
can be found in the manual for each device which can be found on our website:
https://www.spectrum-instrumentation.com/en/downloads

Further information can be found online in the Knowledge Base:
https://www.spectrum-instrumentation.com/en/knowledge-base-overview

**************************************************************************
*/


// ----- include standard driver header from library -----
#include "../c_header/dlltyp.h"
#include "../c_header/regs.h"
#include "../c_header/spcerr.h"
#include "../c_header/spcm_drv.h"

#include "../common/ostools/spcm_oswrap.h"
#include "../common/ostools/spcm_ostools.h"

// ----- standard c include files -----
#include <cstdio>
#include <cstdlib>
#define _USE_MATH_DEFINES
#include <cmath>

#define WAVEFORM_LENGTH (1024*1024)


/*
**************************************************************************
main
**************************************************************************
*/

int main ()
    {
    drv_handle  hCard = NULL_HANDLE;
    char        szErrorTextBuffer[ERRORTEXTLEN], szName[50];

    int64       llSamplerate =  MEGA(50);

    // ------------------------------------------------------------------------
    // uncomment the second line and replace the IP address to use remote
    // cards like in a generatorNETBOX
    int32 lCardIdx = 0;
    sprintf (szName, "/dev/spcm%d", lCardIdx);
    // sprintf (szName, "TCPIP::192.168.1.10::inst%d::INSTR", lCardIdx);
    hCard = spcm_hOpen (szName);

    // no card found
    if (!hCard)
        {
        printf ("no card found...\n");
        return 0;
        }

    // read card type name
    char acCardType[20] = {};
    spcm_dwGetParam_ptr (hCard, SPC_PCITYP, acCardType, sizeof (acCardType));

    // read out some info and print it
    int32 lCardType = 0;
    spcm_dwGetParam_i32 (hCard, SPC_PCITYP,            &lCardType);
    int32 lSerialNumber = 0;
    spcm_dwGetParam_i32 (hCard, SPC_PCISERIALNO,       &lSerialNumber);
    int32 lFncType = 0;
    spcm_dwGetParam_i32 (hCard, SPC_FNCTYPE,           &lFncType);

    switch (lFncType)
        {
        case SPCM_TYPE_AO:
            printf ("Found: %s sn %05d\n", acCardType, lSerialNumber);
            break;

        // only D/A cards allowed in example
        default:
            printf ("Card: %s sn %05d not supported by example\n", acCardType, lSerialNumber);
            return EXIT_FAILURE;
        }

    int32 lMaxDACValue = 0;
    spcm_dwGetParam_i32 (hCard, SPC_MIINST_MAXADCVALUE, &lMaxDACValue);

    int32 lNumModules = 0;
    spcm_dwGetParam_i32 (hCard, SPC_MIINST_MODULES, &lNumModules);
    int32 lNumChPerModule = 0;
    spcm_dwGetParam_i32 (hCard, SPC_MIINST_CHPERMODULE, &lNumChPerModule);
    int32 lNumCh = lNumModules * lNumChPerModule;

    // ------------------------------------------------------------------------
    // do a simple standard setup for all cards
    spcm_dwSetParam_i32 (hCard, SPC_CHENABLE,        (0x1 << lNumCh) - 1);   // enable all channels
    spcm_dwSetParam_i32 (hCard, SPC_CARDMODE,        SPC_REP_STD_SINGLE);    // standard single mode
    spcm_dwSetParam_i32 (hCard, SPC_CLOCKMODE,       SPC_CM_INTPLL);         // clock mode internal PLL
    spcm_dwSetParam_i64 (hCard, SPC_SAMPLERATE,      llSamplerate);          // sampling clock
    spcm_dwSetParam_i32 (hCard, SPC_CLOCKOUT,        0);                     // no clock output
    spcm_dwSetParam_i32 (hCard, SPC_MEMSIZE,         WAVEFORM_LENGTH);       // number of samples per channel
    spcm_dwSetParam_i32 (hCard, SPC_LOOPS,           0);                     // loop continuous

    // trigger mode definition (in our example it is software trigger)
    spcm_dwSetParam_i32 (hCard, SPC_TRIG_ORMASK,     SPC_TMASK_SOFTWARE);
    spcm_dwSetParam_i32 (hCard, SPC_TRIG_ANDMASK,    0);

    // setup the channels
    for (int lChIdx = 0; lChIdx < lNumCh; ++lChIdx)
        {
        spcm_dwSetParam_i32 (hCard, SPC_AMP0       + 100 * lChIdx,   1000); // +-1V output amplitude
        spcm_dwSetParam_i32 (hCard, SPC_ENABLEOUT0 + 100 * lChIdx,      1); // enable output of channel
        }

    // ------------------------------------------------------------------------
    // calculate waveforms
    int64 llBufferSize = sizeof (int16) * lNumCh * (uint64)WAVEFORM_LENGTH;
    int16* anDMABuffer = (int16*)pvAllocMemPageAligned (llBufferSize);
    for (int lChIdx = 0; lChIdx < lNumCh; ++lChIdx)
        {
        // calculate each sample and put it into the DMA buffer (multiplexed)
        for (int i = 0; i < WAVEFORM_LENGTH; ++i)
            {
            int16 nValue = 0;
            switch (lChIdx)
                {
                default:
                case 0:
                    // sine
                    nValue = (int16)(lMaxDACValue * sin (2. * M_PI * i / (WAVEFORM_LENGTH / (lChIdx + 1)))); // for channel zero we generate one period. for channels higher than 3 a faster sine is generated
                    break;
                case 1:
                    // one rectangle period (max amplitude)
                    if (i < WAVEFORM_LENGTH / 2)
                        nValue = (int16)lMaxDACValue - 1;
                    else
                        nValue = (int16)-lMaxDACValue;
                    break;
                case 2:
                    // rising ramp
                    nValue = (int16)(-lMaxDACValue + 2*lMaxDACValue * ((double)i / WAVEFORM_LENGTH));
                    break;
                case 3:
                    {
                    // sinc function in the interval -4pi to +4pi
                    double dX = -4. * M_PI + 8. * M_PI * ((double)i / WAVEFORM_LENGTH);
                    nValue = (int16)(lMaxDACValue * sin (dX) / dX);
                    break;
                    }
                }

            anDMABuffer[lNumCh * i + lChIdx] = nValue;
            }
        }

    printf ("Starting the DMA transfer and waiting until data is in board memory\n");
    spcm_dwDefTransfer_i64 (hCard, SPCM_BUF_DATA, SPCM_DIR_PCTOCARD, 0, anDMABuffer, 0, llBufferSize);
    spcm_dwSetParam_i64 (hCard, SPC_DATA_AVAIL_CARD_LEN, llBufferSize);
    spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_DATA_STARTDMA | M2CMD_DATA_WAITDMA);
    printf ("... data has been transferred to board memory\n");

    // We'll start and wait until the card has finished or until a timeout occurs
    spcm_dwSetParam_i32 (hCard, SPC_TIMEOUT, 10 * 1000); // 10 seconds timeout
    printf ("\nStarting the card and waiting for ready interrupt\n(continuous and single restart will have timeout)\n");
    // start command including enable of trigger engine -> error check
    if (spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_START | M2CMD_CARD_ENABLETRIGGER | M2CMD_CARD_WAITREADY) != ERR_OK)
        {
        spcm_dwGetErrorInfo_i32 (hCard, NULL, NULL, szErrorTextBuffer);
        printf ("Error: %s\n", szErrorTextBuffer);
        }

    printf ("\nFinished ...\n");

    // send the stop command
    spcm_dwSetParam_i32 (hCard, SPC_M2CMD, M2CMD_CARD_STOP | M2CMD_DATA_STOPDMA);

    // clean up
    vFreeMemPageAligned (anDMABuffer, llBufferSize);
        
    spcm_vClose (hCard);

    return EXIT_SUCCESS;
    }
